/* "$Id: xapmmon.c,v 1.25 2004/06/02 01:55:42 kiesling Exp $" */

/*
  This file is part of xapmmon
  Copyright  2004 Robert Kiesling, rkies@cpan.org.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>
#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "xpm/bgmask.xpm"
#include "xpm/bar.xpm"

extern int errno;

void update (int);
int apm_time (void);
int read_apm (void);
void help (void);
int parse_args (int , char **);
void app_exit (int);
void background_pixmap (void);
void bar_pixmap (void);
int get_screen_palette (void);

Display *display;
int screen, screen_depth, x_connection_no;
Window root, win;
GC gc;
Pixmap background, background_mask;
Pixmap bar, bar_mask;
XpmAttributes background_attributes;
XpmAttributes bar_attributes;
static char bgcolor[255] ="lightskyblue";
static char fgcolor[255] = "white";
static char bevel1color[255] = "slategray";
static char bevel2color[255] = "lightgray";
static char bevel3color[255] = "black";
static char bevel4color[255] = "whitesmoke";
XColor bgXcolor, fgXcolor, b1Xcolor, b2Xcolor, b3Xcolor, b4Xcolor, *palette;
unsigned int n_colors;

int x_org = 10, y_org = 10;

static char app_name[255];

int poll_interval = 0.5;

#define APM_FILE "/proc/apm"

int main (int argc, char **argv) {

  int pcnt;
  char *d_env = NULL;
  XGCValues gcv;
  unsigned long gcv_mask;
  XEvent e;
  XWindowAttributes window_attributes;
  XSetWindowAttributes set_attributes;
  XWindowChanges config;
  XClassHint classhint;
  int wm_event_mask = ExposureMask | ClientMessage | DestroyNotify;
  Atom wm_delete_window;

  parse_args (argc, argv);

  pcnt = read_apm ();

  if ((display = XOpenDisplay (d_env)) == NULL) {
    fprintf (stderr, "%s: can't open display %s.\n", 
	     app_name, XDisplayName(d_env));
    exit (1);
  }

  screen = DefaultScreen (display);
  root = RootWindow (display, screen);
  screen_depth = DefaultDepth (display, screen);

  get_screen_palette ();

  XGetWindowAttributes (display, root, &window_attributes);

  win = XCreateWindow (display, root, 0, 0, 104, 25, 0, screen_depth,
		       CopyFromParent, CopyFromParent, 0, &set_attributes);

  wm_delete_window = XInternAtom (display, "WM_DELETE_WINDOW", False);
  XSetWMProtocols (display, win, &wm_delete_window, 1);
  
  classhint.res_name = "xapmmon";
  classhint.res_class = "Xapmmon";
  XSetClassHint (display, win, &classhint);

  background_pixmap ();
  bar_pixmap ();

  XSetWindowBackgroundPixmap (display, win, background);

  XSelectInput(display, win, wm_event_mask);

  gcv_mask = GCFunction | GCFillStyle;
  gcv.fill_style = FillSolid;
  gcv.function = GXcopy;

  gc = XCreateGC (display, win, gcv_mask, &gcv);

  XMapWindow (display, win);
  XRaiseWindow (display, win);

  config.x = x_org;
  config.y = y_org;
  XConfigureWindow (display, win, CWX | CWY, &config);
		     
  update (pcnt);

  while (1) {
    while (XPending(display)) {
      XNextEvent (display, &e);
      switch (e.type)
	{
	case Expose:
	  pcnt = read_apm ();
	  update (pcnt);
	  break;
	case DestroyNotify:
	  app_exit (0);
	  break;
	case ClientMessage:
	  if(e.xclient.data.l[0] == wm_delete_window)
	    app_exit (0);
	default:
	  break;
	}
      XFlush (display);
    }
    pcnt = read_apm ();
    update (pcnt);
    usleep (poll_interval * 1000000L);
  }

  app_exit (0);

  return 0;
}

get_screen_palette (void) {

  int i;
  Colormap default_cmap;

  n_colors = XDisplayCells (display, screen);

  palette = (XColor *)calloc (n_colors, sizeof(XColor));
  
  for (i = 0; i < n_colors; i++)
    palette[i].pixel = i;

  default_cmap = DefaultColormap (display, screen);
  XQueryColors (display, default_cmap, palette, n_colors);

  return n_colors;
}

/* The method of color selection used by GNU Emacs and GDK is to 
select an already allocated color from the screen palette using a 
least-squares fit with the requested color. */
int lookup_color (XColor *color, char *name) {

  XColor exact_color, screen_color;
  Colormap default_cmap;
  int i, r;
  unsigned long var, fit, match, r_param, b_param, g_param;

  default_cmap = DefaultColormap (display, screen);

  r = XLookupColor (display, default_cmap, name, &exact_color, &screen_color);

  fit = 0xFFFF;
  for (i = 0; i < n_colors; i++) {
    r_param = (exact_color.red >> 8) - (palette[i].red >> 8);
    b_param = (exact_color.blue >> 8) - (palette[i].blue >> 8);
    g_param = (exact_color.green >> 8)- (palette[i].green >> 8);

    var = r_param * r_param + b_param * b_param + g_param * g_param;
    if (var < fit) {
      fit = var;
      match = i;
    }
  }

  color -> pixel = palette[match].pixel;
  color -> red = palette[match].red;
  color -> green = palette[match].green;
  color -> blue = palette[match].blue;
  color -> flags = palette[match].flags;
  
  return 0;
}

int rgb_string (char *buf, XColor color) {

  sprintf (buf, "#%02X%02X%02X",
	   color.red >> 8,
	   color.green >> 8,
	   color.blue >> 8);

  return 0;
}

void background_pixmap (void) {

  int r;
  char d1[255], d2[255], d3[255], d4[255], c1[255], c2[255], c3[255], c4[255];

  lookup_color (&bgXcolor, bgcolor);
  lookup_color (&b1Xcolor, bevel1color);
  lookup_color (&b2Xcolor, bevel2color);
  lookup_color (&b3Xcolor, bevel3color);

  rgb_string (c1, bgXcolor);
  rgb_string (c2, b1Xcolor);
  rgb_string (c3, b2Xcolor);
  rgb_string (c4, b3Xcolor);

  sprintf (d1, ". c %s", c1);
  sprintf (d2, "b c %s", c2);
  sprintf (d3, "l c %s", c3);
  sprintf (d4, "m c %s", c4);

  bgdata[1] = &d1[0];
  bgdata[2] = &d2[0];
  bgdata[3] = &d3[0];
  bgdata[4] = &d4[0];

  if ((r = XpmCreatePixmapFromData (display, root, bgdata, &background,
				    &background_mask, 
				    &background_attributes)) != XpmSuccess) {
    switch (r)
      {
      case 1:
	fprintf (stderr, "%s: Invalid color in pixmap.\n", app_name);
	break;
      case -1:
	fprintf (stderr, "%s: Couldn't open pixmap file.\n", app_name);
	break;
      case -2:
	fprintf (stderr, "%s: Bad pixmap file.\n", app_name);
	break;
      case -3:
	fprintf (stderr, "%s: Pixmap memory error.\n", app_name);
	break;
      case -4:
	fprintf (stderr, "%s: Couldn't create pixmap color %s.\n", 
		 app_name, bgcolor);
	break;
      }
    exit (r);
  }
}

void bar_pixmap (void) {

  int r;
  char d1[255], d2[255], d3[255], c1[255], c2[255], c3[255];

  lookup_color (&fgXcolor, fgcolor);
  lookup_color (&b1Xcolor, bevel1color);
  lookup_color (&b4Xcolor, bevel4color);

  rgb_string (c1, fgXcolor);
  rgb_string (c2, b4Xcolor);
  rgb_string (c3, b1Xcolor);

  sprintf (d1, ". c %s", c1);
  sprintf (d2, "a c %s", c2);
  sprintf (d3, "b c %s", c3);

  bar_data[1] = &d1[0];
  bar_data[2] = &d2[0];
  bar_data[3] = &d3[0];

  if ((r = XpmCreatePixmapFromData (display, root, bar_data, &bar, 
			       &bar_mask, 
			       &bar_attributes)) != XpmSuccess) {
    switch (r)
      {
      case 1:
	fprintf (stderr, "%s: Invalid color in pixmap.\n", app_name);
	break;
      case -1:
	fprintf (stderr, "%s: Couldn't open pixmap file.\n", app_name);
	break;
      case -2:
	fprintf (stderr, "%s: Bad pixmap file.\n", app_name);
	break;
      case -3:
	fprintf (stderr, "%s: Pixmap memory error.\n", app_name);
	break;
      case -4:
	fprintf (stderr, "%s: Couldn't create pixmap color %s.\n", 
		 app_name, fgcolor);
	break;
      }
    exit (r);
  }
}

void update (int percent) {

  XClearArea (display, win, 1, 1 + percent, 100 - (1 + percent), 20, 0);
  XCopyArea (display, background, win, gc, 0,0, 0, 0, 0,0);
  XCopyArea (display, bar, win, gc, 0,0, 2, 22, 2,1);
  XCopyArea (display, bar, win, gc, 100 - percent,0, 100, 22, 3,1);
}

int apm_time (void) {
  struct timeval tv;
  struct timezone tz;

  gettimeofday (&tv, &tz);

  return tv.tv_sec;
}

int read_apm (void) {
  
  FILE *f;
  char *p, *q, buf[255];
  
  if ((f = fopen (APM_FILE, "r")) == NULL) {
    fprintf (stderr, "%s: %s\n", APM_FILE, strerror (errno));
    app_exit (1);
  }
  
  fgets (buf, 255, f);

  fclose (f);

  q = NULL;
  q = strchr (buf, '%');
  if (!q) {
    fprintf (stderr, "%s: unable to read %s stats.\n", app_name, APM_FILE);
    app_exit (1);
  }
  *q = 0;
  p = rindex (buf, ' ');
  return atoi (p);
}

int parse_args (int c, char **a) {

  int i;

  strcpy (app_name, a[0]);

  for (i = 1; i < c; i++) {
    if (a[i][0] == '-') {
      switch (a[i][1]) 
	{
	case 'i':
	  poll_interval = atoi(&a[i][2]);
	  break;
	case 'x':
	  x_org = atoi(&a[i][2]);
	  break;
	case 'y':
	  y_org = atoi(&a[i][2]);
	  break;
	case 'b':
	  strcpy (bgcolor, &a[i][3]);
	  break;
	case 'f':
	  strcpy (fgcolor, &a[i][3]);
	  break;
	case 'h':
	default:
	  help ();
	  break;
	}
    }
  }
  return 0;
}

void help (void) {
  fprintf (stderr, 
	   "Usage: %s [-h] | [-bg<color>] [-fg<color>] [-i<interval>] [-x<pixels>] [-y<pixels>]\n", 
	   app_name);
  fprintf (stderr, "-bg<color>       Background color.\n");
  fprintf (stderr, "-fg<color>       Foreground color.\n");
  fprintf (stderr, "-h               Print this message and exit.\n");
  fprintf (stderr, "-i<interval>     Interval in seconds to check power level.\n");
  fprintf (stderr, "-x<pixels>       X coordinate of window's upper left hand corner.\n");
  fprintf (stderr, "-y<pixels>       Y coordinate of window's upper left hand corner.\n");
  exit (1);
}

void app_exit (int r) {

  XSetCloseDownMode (display, DestroyAll);
  XFreeGC (display, gc);
/*   XFreePixmap (display, background); */
/*   XFreePixmap (display, background_mask); */
/*   XFreePixmap (display, bar_mask); */
/*   XFreePixmap (display, bar); */
  XDestroyWindow (display, win);
  XCloseDisplay (display);
  exit (r);
}

